/*--------------------------------------------------------------------*/
/*                                                                    */
/*  HALLOCTR.H                                                        */
/*                                                                    */
/*--------------------------------------------------------------------*/

#if !defined( HALLOCTR_H )
#define HALLOCTR_H

#if defined(__DPMI32__) || defined(__WIN32__) || defined(__OS2__)
#error Can not use huge allocation in a 32 bit model
#endif

#if !defined(___DEFS_H)
#include <_defs.h>
#endif

#if !defined( __STDLIB_H )
#include <stdlib.h>
#endif

#if !defined( CLASSLIB_DEFS_H )
#include <classlib/defs.h>
#endif

#if !defined(__NEW_H)
#include <new.h>
#endif

#if defined( BI_CLASSLIB_NO_po )
#pragma option -po-
#endif

/*------------------------------------------------------------------------*/
/*                                                                        */
/*  class THugeStandardAllocator                                          */
/*                                                                        */
/*  Provides class-specific operator new and operator delete that         */
/*  simply call the global operator new and operator delete.  That        */
/*  is, THugeStandardAllocator does not provide any specialized behavior. */
/*  It is used in the non-managed versions of the parametrized            */
/*  containers.                                                           */
/*                                                                        */
/*------------------------------------------------------------------------*/

class _BIDSCLASS THugeStandardAllocator
{
public:
    friend void *
            operator new( unsigned long sz, const THugeStandardAllocator& )
        { return ::operator new( sz ); }
    friend void *
            operator new [] ( unsigned long sz, const THugeStandardAllocator& )
        { return ::operator new [] ( sz ); }
    friend void *
         operator new( unsigned long, void __huge *ptr )
        { return ptr; }
    friend void *
         operator new [] ( unsigned long, void __huge *ptr )
        { return ptr; }
    void operator delete( void __huge *ptr )
        { ::operator delete( ptr ); }
    void operator delete [] ( void __huge *ptr )
        { ::operator delete [] ( ptr ); }
};

/*------------------------------------------------------------------------*/
/*                                                                        */
/*  class THugeSharedAllocator                                            */
/*                                                                        */
/*  Provides class-specific operator new and operator delete that         */
/*  allocate from shared memory.                                          */
/*                                                                        */
/*------------------------------------------------------------------------*/

class _BIDSCLASS THugeSharedAllocator
{
public:
    friend void *operator new( unsigned long sz, const THugeSharedAllocator& );
    friend void *operator new [] ( unsigned long sz, const THugeSharedAllocator& );
    void operator delete( void __huge *ptr );
    void operator delete [] ( void __huge *ptr );
};

/*------------------------------------------------------------------------*/
/*                                                                        */
/*  template <class T, class Alloc> class THugeManaged_T                  */
/*                                                                        */
/*  Provides a parametrized wrapper for type T which uses a               */
/*  class-specific operator new and operator delete supplied              */
/*  by Alloc.                                                             */
/*                                                                        */
/*------------------------------------------------------------------------*/

template <class T, class Alloc> class THugeManaged_T : private Alloc
{

public:

    Alloc::operator delete;
    Alloc::operator delete [];

    THugeManaged_T() {}
    THugeManaged_T( const T& t ) : data(t)
        {
        }

    const THugeManaged_T& operator = ( const THugeManaged_T& t )
        {
        data = t.data;
        return *this;
        }

    operator T&()
        {
        return data;
        }

private:

    T data;

};

/*------------------------------------------------------------------------*/
/*                                                                        */
/*  THugeSharedAllocator::operator new                                    */
/*  THugeSharedAllocator::operator delete                                 */
/*                                                                        */
/*  When compiling for WINDOWS, allocates memory as GMEM_DDESHARE.        */
/*  When compiling for DOS, uses the global operator new and              */
/*  operator delete.                                                      */
/*                                                                        */
/*------------------------------------------------------------------------*/

#if defined( _Windows )

#if !defined( __WINDOWS_H )
#include <windows.h>
#endif

#if !defined( __DOS_H )
#include <dos.h>
#endif

inline void *operator new( unsigned long sz,
                                const THugeSharedAllocator& )
{
    return GlobalLock( GlobalAlloc( GPTR | GMEM_DDESHARE, sz ) );
}

inline void *operator new [] ( unsigned long sz,
                                      const THugeSharedAllocator& )
{
    return GlobalLock( GlobalAlloc( GPTR | GMEM_DDESHARE, sz ) );
}

inline void THugeSharedAllocator::operator delete( void __huge *ptr )
{
    HGLOBAL hMem = (HGLOBAL)GlobalHandle(FP_SEG((void __far *)ptr));
    if( GlobalUnlock(hMem) == 0 )
        GlobalFree(hMem);
}

inline void THugeSharedAllocator::operator delete [] ( void __huge *ptr )
{
    HGLOBAL hMem = (HGLOBAL)GlobalHandle(FP_SEG((void __far *)ptr));
    if( GlobalUnlock(hMem) == 0 )
        GlobalFree(hMem);
}

#else

inline void *operator new( unsigned long sz,
                                    const THugeSharedAllocator& )
{
    return ::operator new( sz );
}

inline void *operator new [] ( unsigned long sz,
                                    const THugeSharedAllocator& )
{
    return ::operator new [] ( sz );
}

inline void THugeSharedAllocator::operator delete( void __huge *ptr )
{
    ::operator delete((void __far *) ptr );
}

inline void THugeSharedAllocator::operator delete [] ( void __huge *ptr )
{
    ::operator delete [] ((void __far *) ptr );
}

#endif  // _Windows

#if defined( BI_CLASSLIB_NO_po )
#pragma option -po.
#endif

#endif  // HALLOCTR_H
